@@ -2,7 +2,7 @@
  *  typespeed - measures your typing speed
  *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
  *                          & Jaakko Manelius  <jman@iki.fi>
- *  Copyright (C) 2006-2007   Tobias Stoeckmann  <tobias@bugol.de>
+ *  Copyright (C) 2006-2008   Tobias Stoeckmann  <tobias@bugol.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -351,8 +351,8 @@ compar(const void *p1, const void *p2)
 	elem2 = (struct finfo *)p2;
 
 	if (!(retval = strcmp(elem1->descr, elem2->descr)))
-		xerrx(1, "%s and %s have the same description!\n",
-		    elem1->name, worddir, elem2->name);
+		xerrx(1, "%s/%s and %s/%s have the same description!",
+		    worddir, elem1->name, worddir, elem2->name);
 	return retval;
 }
 
@@ -439,15 +439,15 @@ getentry(char *line)
 		switch(pos) {
 		case 0: /* count */
 			if (!(entry->count = cstrl(line)) && errno)
-				xerrx(1, "count is invalid: %s\n", line);
+				xerrx(1, "count is invalid: %s", line);
 			break;
 		case 1: /* tcount */
 			if (!(entry->tcount = cstrl(line)) && errno)
-				xerrx(1, "tcount is invalid: %s\n", line);
+				xerrx(1, "tcount is invalid: %s", line);
 			break;
 		case 2: /* wordswritten */
 			if (!(entry->wordswritten = cstrl(line)) && errno)
-				xerrx(1, "wordswritten is invalid: %s\n", line);
+				xerrx(1, "wordswritten is invalid: %s", line);
 			break;
 		case 3: /* name */
 			while (*line == ' ')
@@ -455,7 +455,7 @@ getentry(char *line)
 			t = unescstr(line);
 			if (xsnprintf(entry->name, sizeof(entry->name),
 			    "%s", t))
-				xerrx(1, "name is invalid: %s\n", t);
+				xerrx(1, "name is invalid: %s", t);
 			free(t);
 			break;
 		case 4: /* word list */
@@ -464,7 +464,7 @@ getentry(char *line)
 			t = unescstr(line);
 			if (xsnprintf(entry->wordlist, sizeof(entry->wordlist),
 			    "%s", t))
-				xerrx(1, "word list is invalid: %s\n", t);
+				xerrx(1, "word list is invalid: %s", t);
 			free(t);
 			break;
 		case 5: /* mod */
@@ -472,12 +472,12 @@ getentry(char *line)
 				line++;
 			t = unescstr(line);
 			if (xsnprintf(entry->mod, sizeof(entry->mod), "%s", t))
-				xerrx(1, "game rule is invalid: %s\n", t);
+				xerrx(1, "game rule is invalid: %s", t);
 			free(t);
 			break;
 		case 6: /* duration */
 			if (!(entry->duration = cstrl(line)) && errno)
-				xerrx(1, "duration is invalid: %s\n", line);
+				xerrx(1, "duration is invalid: %s", line);
 			break;
 		default:
 			xerrx(1, "too many fields in high score file");
@@ -488,7 +488,7 @@ getentry(char *line)
 	}
 	/* sinit */
 	if (!(entry->sinit = cstrl(line)) && errno)
-		xerrx(1, "sinit is invalid: %s\n", line);
+		xerrx(1, "sinit is invalid: %s", line);
 
 	return entry;
 }
@@ -622,59 +622,71 @@ getnum(char *p)
 int
 hcompar(const void *entry1, const void *entry2)
 {
+	int i;
 	struct hentry **en1, **en2;
 	unsigned long val1, val2;
 	float rat1, rat2;
 
 	en1 = (struct hentry **)entry1;
 	en2 = (struct hentry **)entry2;
-	val1 = (*en1)->count;
-	val2 = (*en2)->count;
 
-	if (val1 < val2)
-		return 1;
-	if (val1 == val2) {
-		if ((*en1)->duration)
-			rat1 = (float)((*en1)->count +
-			    (*en1)->wordswritten) /
-			    (*en1)->duration * 100.0f;
-		else
-			rat1 = 0.0f;
+	for (i = 0; i < 3; i++) {
+		switch (opt.order[i]) {
+		case 'c':
+			if ((*en1)->duration)
+				rat1 = (float)((*en1)->count +
+				    (*en1)->wordswritten) /
+				    (*en1)->duration;
+			else
+				rat1 = 0.0f;
 
-		if ((*en2)->duration)
-			rat2 = (float)((*en2)->count +
-			    (*en2)->wordswritten) /
-			    (*en2)->duration * 100.0f;
-		else
-			rat2 = 0.0f;
+			if ((*en2)->duration)
+				rat2 = (float)((*en2)->count +
+				    (*en2)->wordswritten) /
+				    (*en2)->duration;
+			else
+				rat2 = 0.0f;
 
-		if (rat1 < rat2)
-			return 1;
-		if (rat1 > rat2)
-			return -1;
+			if (rat1 < rat2)
+				return 1;
+			if (rat1 > rat2)
+				return -1;
+			break;
+		case 's':
+			val1 = (*en1)->count;
+			val2 = (*en2)->count;
+			if (val1 < val2)
+				return 1;
+			else if (val1 > val2)
+				return -1;
+			break;
+		case 't':
+			if (!(*en1)->tcount)
+				rat1 = 0.0f;
+			else
+				rat1 = (1 - (float)((*en1)->count +
+				    (*en1)->wordswritten) / ((*en1)->tcount +
+				    (*en1)->wordswritten));
 
-		if ((*en1)->duration)
-			rat1 = (float)((*en1)->tcount +
-			    (*en1)->wordswritten) /
-			    (*en1)->duration * 100.0f;
-		else
-			rat1 = 0.0f;
+			if (!(*en2)->tcount)
+				rat2 = 0.0f;
+			else
+				rat2 = (1 - (float)((*en2)->count +
+				    (*en2)->wordswritten) / ((*en2)->tcount +
+				    (*en2)->wordswritten));
 
-		if ((*en2)->duration)
-			rat2 = (float)((*en2)->tcount +
-			    (*en2)->wordswritten) /
-			    (*en2)->duration * 100.0f;
-		else
-			rat2 = 0.0f;
-
-		if (rat1 > rat2)
-			return 1;
-		if (rat1 < rat2)
-			return -1;
-
-		return 0;
+			if (rat1 > rat2)
+				return 1;
+			if (rat1 < rat2)
+				return -1;
+			break;
+		default:
+			xerrx(1, "%c in highorder is invalid", opt.order[i]);
+			/* NOTREACHED */
+		}
 	}
-	return -1;
+
+	return 0;
 }
 
 /*
@@ -1010,7 +1022,8 @@ loadwords(char *filename)
 		if (words.n >= words.max) {
 			if ((size_t)-1 - 1024 < words.max)
 				xerrx(1, "loadwords: too many words!");
-			pointer = realloc(words.word, words.max + 1024);
+			pointer = realloc(words.word,
+			    (words.max + 1024) * sizeof(char *));
 			if (pointer == NULL)
 				xerr(1, NULL);
 			words.word = pointer;
@@ -1186,7 +1199,7 @@ readfile(char *filename, int op)
 			line[sizeof(line) - 3] = '\0';
 			strncat(line, "\n", 1);
 			if (parseline(line, &option, &value))
-				xerrx(1, "Error in %s line %lu\n", filename,
+				xerrx(1, "Error in %s line %lu", filename,
 				    lineno);
 			if (option != NULL && value != NULL)
 				setoptions(option, value, op);
@@ -1198,7 +1211,7 @@ readfile(char *filename, int op)
 		while (fgets(line, sizeof(line), file) != NULL) {
 			lineno++;
 			if (parseline(line, &option, &value))
-				xerrx(1, "Error in %s line %lu\n", filename,
+				xerrx(1, "Error in %s line %lu", filename,
 				    lineno);
 			if (option != NULL && value != NULL)
 				setoptions(option, value, op);
@@ -1209,6 +1222,34 @@ readfile(char *filename, int op)
 	}
 }
 
+char *
+xstrsep(char **str, char *delim)
+{
+	char *p, *ret;
+
+	if (*str == NULL)
+		return NULL;
+
+	ret = *str;
+	for (p = *str; *p != '\0'; p++)
+		if (strchr(delim, *p) != NULL)
+			break;
+	if (*p == '\0') {
+		*str = NULL;
+		return ret;
+	}
+
+	for (*p++ = '\0'; *p != '\0' && strchr(delim, *p) != NULL; p++)
+		;
+	if (*p == '\0') {
+		*str = NULL;
+		return ret;
+	}
+
+	*str = p;
+	return ret;
+}
+
 /*
  * setoptions sets actually configuration and rule sets. If a
  * configuration or rule set has been opened decides "op":
@@ -1218,6 +1259,8 @@ readfile(char *filename, int op)
 static void
 setoptions(char *option, char *value, int op)
 {
+	int i;
+	char *p;
 	unsigned long val;
 
 	if (op) { /* configuration file */
@@ -1227,7 +1270,31 @@ setoptions(char *option, char *value, int op)
 			else
 				opt.cheat = 0;
 		}
-		else if (!strcmp(option, "ruledir"))
+		else if (!strcmp(option, "highorder")) {
+			for (i = 0; i < 3; i++) {
+				if ((p = xstrsep(&value, " \t")) == NULL) {
+					printf("break\n");
+					break;
+				}
+
+				switch (p[0]) {
+				case 'c':
+				case 's':
+				case 't':
+					opt.order[i] = p[0];
+					break;
+				default:
+					printf("def\n");
+					xerrx(1, "highorder: %s unknown", p);
+					/* NOTREACHED */
+				}
+			}
+			if (i != 3 || p == NULL ||
+			    (p = xstrsep(&value, " \t")) != NULL) {
+				printf("abc\n");
+				xerrx(1, "highorder invalid");
+			}
+		} else if (!strcmp(option, "ruledir"))
 			xstrncpy(ruledir, value, sizeof(ruledir) - 1);
 		else if (!strcmp(option, "worddir"))
 			xstrncpy(worddir, value, sizeof(worddir) - 1);
@@ -1237,33 +1304,33 @@ setoptions(char *option, char *value, int op)
 	else { /* game rule */
 		if (!strcmp(option, "misses")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "misses invalid: %s\n", value);
+				xerrx(1, "misses invalid: %s", value);
 			if (val > 0 && val < 100)
 				rules.misses = val;
 		}
 		else if (!strcmp(option, "min word length")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "min word length invalid: %s\n",
+				xerrx(1, "min word length invalid: %s",
 				    value);
 			if (val > 0 && val < 20)
 				rules.minlen = val;
 		}
 		else if (!strcmp(option, "max word length")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "max word length invalid: %s\n",
+				xerrx(1, "max word length invalid: %s",
 				    value);
 			if (val > 0 && val < 20)
 				rules.maxlen = val;
 		}
 		else if (!strcmp(option, "min words")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "min words invalid: %s\n", value);
+				xerrx(1, "min words invalid: %s", value);
 			if (val > 0 && val < 23)
 				rules.minwords = val;
 		}
 		else if (!strcmp(option, "max words")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "max words invalid: %s\n", value);
+				xerrx(1, "max words invalid: %s", value);
 			if (val > 0 && val < 23)
 				rules.maxwords = val;
 		}
@@ -1275,18 +1342,18 @@ setoptions(char *option, char *value, int op)
 		}
 		else if (!strcmp(option, "min score")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "min score invalid: %s\n", value);
+				xerrx(1, "min score invalid: %s", value);
 			rules.minscore = val;
 		}
 		else if (!strcmp(option, "min speed")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "min speed invalid: %s\n", value);
+				xerrx(1, "min speed invalid: %s", value);
 			if (val < 100)
 				rules.minspeed = val;
 		}
 		else if (!strcmp(option, "max speed")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "max speed invalid: %s\n", value);
+				xerrx(1, "max speed invalid: %s", value);
 			if (val < 100)
 				rules.maxspeed = val;
 		}
@@ -1294,7 +1361,7 @@ setoptions(char *option, char *value, int op)
 			xstrncpy(rules.name, value, sizeof(rules.name) - 1);
 		else if (!strcmp(option, "step")) {
 			if (!(val = cstrl(value)) && errno)
-				xerrx(1, "step invalid: %s\n", value);
+				xerrx(1, "step invalid: %s", value);
 			if (val < 1000)
 				rules.step = val;
 		}
