@@ -2,7 +2,7 @@
  *  typespeed - measures your typing speed
  *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
  *                          & Jaakko Manelius  <jman@iki.fi>
- *  Copyright (C) 2006-2007   Tobias Stoeckmann  <tobias@bugol.de>
+ *  Copyright (C) 2006-2008   Tobias Stoeckmann  <tobias@bugol.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -129,7 +129,7 @@ void			win32_sighandler(int);
 #ifndef TEST
 static void		clearword(int, int, size_t);
 static void		cwords(void);
-static void		movewords(void);
+static void		movewords(int *);
 static void		parseinput(int *, char *, clock_t *, unsigned *, int *, int*);
 static int		play(void);
 static void		usage(void);
@@ -547,23 +547,28 @@ Typorank:\t"),
 #ifndef TEST
 /*
  * Move words across the screen.
- * Removes them if they reach the right side and increases misses.
+ * Removes them if they reach the right side and returns the amount of
+ * words which left the screen.
  */
 static void
-movewords(void)
+movewords(int *misses)
 {
-	int i, length;
+	int buf[4][22]; 
+	int count[4];
+	int i, j, length;
 
+	memset(count, 0, sizeof(count));
+
+	xcolor_set(0);
 	for (i = 0; i < 22; i++) {
 		if (wordpos[i] == -2)
 			continue;
 
 		length = (int)strlen(wordstring[i]);
-		xcolor_set(0);
 
 		if (wordpos[i] > (79 - length)) {
 			clearword(i, wordpos[i], length);
-			misses++;
+			buf[3][count[3]++] = i;
 			wordpos[i] = -2;
 			continue;
 		}
@@ -572,13 +577,46 @@ movewords(void)
 			mvaddch(i, wordpos[i], ' ');
 
 		if (wordpos[i] > (65 - length))
-			xcolor_set(3);
+			buf[2][count[2]++] = i;
 		else if (wordpos[i] > (50 - length))
-			xcolor_set(7);
+			buf[1][count[1]++] = i;
 		else
-			xcolor_set(1);
+			buf[0][count[0]++] = i;
+	}
 
-		mvaddstr(i, ++wordpos[i], wordstring[i]);
+	*misses += count[3];
+
+	/*
+	 * Get words that are missed back in buffer, maybe the player
+	 * was about to type them.
+	 */
+	if (*misses >= rules.misses) {
+		for (i = 0; i < count[3]; i++)
+			wordpos[buf[3][i]] = 0;
+		return;
+	}
+
+	for (j = 0; j < 3; j++) {
+		if (!count[j])
+			continue;
+
+		switch (j) {
+		case 0:
+			xcolor_set(1);
+			break;
+		case 1:
+			xcolor_set(7);
+			break;
+		case 2:
+			xcolor_set(3);
+			break;
+		default:
+			break;
+		}
+
+		for (i = 0; i < count[j]; i++)
+			mvaddstr(buf[j][i], ++wordpos[buf[j][i]],
+			    wordstring[buf[j][i]]);
 	}
 }
 #endif /* TEST */
@@ -809,7 +847,21 @@ play(void)
 		oldtimes = timenow();
 		now.duration = oldtimes - starttime;
 
-		movewords();
+		movewords(&misses);
+
+		/*
+		 * Be nice and check if a player was to typo something while
+		 * the last word slipped away... Every other situation is a
+		 * typo though.
+		 */
+		if (misses >= rules.misses)
+			for (i = 0; i < 22; i++)
+				if (wordpos[i] != -2 &&
+				    !strncmp(wordstring[i], input, inputpos)) {
+					now.score += inputpos;
+					break;
+				}
+
 		cwords();
 
 		now.speed = (now.score + now.wordswritten) * 100.0 /
